structure Dynamics:> DYNAMICS = struct
  datatype d =
    Step of Expr.t
  | Val

  exception Malformed

  fun trystep (e: Expr.t): d =
    case e of
      Expr.Var _ => raise Malformed
    | Expr.Int _ => Val
    | Expr.Str _ => Val
    | Expr.Plus (e1, e2) => (
        case trystep e1 of
          Step e1 => Step (Expr.Plus (e1, e2))
        | Val => (
            case trystep e2 of
              Step e2 => Step (Expr.Plus (e1, e2))
            | Val => (
                case (e1, e2) of
                  (Expr.Int n1, Expr.Int n2) => Step (Expr.Int (n1 + n2))
                | _ => raise Malformed
              )
          )
      )
    | Expr.Concat (e1, e2) => (
        case trystep e1 of
          Step e1 => Step (Expr.Concat (e1, e2))
        | Val => (
            case trystep e2 of
              Step e2 => Step (Expr.Concat (e1, e2))
            | Val => (
                case (e1, e2) of
                  (Expr.Str n1, Expr.Str n2) => Step (Expr.Str (n1 ^ n2))
                | _ => raise Malformed
              )
          )
      )
    | Expr.Len e1 => (
        case trystep e1 of
          Step e1 => Step (Expr.Len e1)
        | Val => (
            case e1 of
              Expr.Str s => Step (Expr.Int (String.size s))
            | _ => raise Malformed
          )
      )
    | Expr.ToStr e1 => (
        case trystep e1 of
          Step e1 => Step (Expr.ToStr e1)
        | Val => (
            case e1 of
              Expr.Int n => Step (Expr.Str (Int.toString n))
            | _ => raise Malformed
          )
      )
    | Expr.Let (x, e1, e2) => (
        case trystep e1 of
          Step e1 => Step (Expr.Let (x, e1, e2))
          (* will not incur capture, since e1 is a value which means it has no
           * free variables (in our simple language, where the only values are
           * ints and strings).
           *)
        | Val => Step (Expr.subst e1 x e2)
      )

  fun eval (e: Expr.t): Expr.t =
    case trystep e of
      Step e => eval e
    | Val => e

  fun synthtypeAndEval (e: Expr.t): (Expr.t * Typ.t) option =
    case Statics.synthtype (Ctx.empty ()) e of
      SOME t => SOME (eval e, t)
    | NONE => NONE
end
