structure RunTests = struct
  type out = (Expr.t * Typ.t) option

  fun outToString (x: out): string =
    case x of
      NONE => "NONE"
    | SOME (e, t) => "SOME (" ^ Expr.toString e ^ ", " ^ Typ.toString t ^ ")"

  fun println (x: string): unit = print (x ^ "\n")

  fun test ((e, expect): Expr.t * out): bool =
    let
      val actual = SOME (Dynamics.synthtypeAndEval e) handle _ => NONE
      val ok =
        case actual of
          NONE => false
        | SOME a => expect = a
      val () = if ok then () else println (String.concatWith "\n"
        [ "fail"
        , "  input: " ^ Expr.toString e
        , "  expect: " ^ outToString expect
        , "  actual: " ^ (
          case actual of
            NONE => "<raise>"
          | SOME a => outToString a
        )
        ])
    in
      ok
    end

  val total = List.length Tests.synthtypeAndEval
  val pass = List.length (List.filter test Tests.synthtypeAndEval)

  val () = println (Int.toString pass ^ "/" ^ Int.toString total ^ " passed")
  val () =
    if OS.FileSys.getDir () = "/home/autolab" then
      println ("{\"scores\":{\"tests\":" ^ Int.toString pass ^ "}}")
    else
      ()
  val () = OS.Process.exit OS.Process.success
end
